/*
 *	HID controller 
 *
 *	programmed by piro
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */
#include "hidcontroller.h"
#include <signal.h>

#include <qpe/qpeapplication.h>
#include <qpe/qcopenvelope_qws.h>
#include <qpe/config.h>

#include <qpushbutton.h>
#include <qbuttongroup.h>
#include <qradiobutton.h>
#include <qwidget.h>
#include <qlabel.h>
#include <qlineedit.h>
#include <qlcdnumber.h>
#include <qspinbox.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <qslider.h>
#include <qfile.h>
#include <qtextstream.h>

#define MOUSE_L	0x01
#define MOUSE_R 0x02

#define CTRL	0x01
#define SHIFT	0x02
#define ALT	0x04

#include "keydata.h" 

/* 
 *  Constructs a Example which is a child of 'parent', with the 
 *  name 'name' and widget flags set to 'f' 
 */
hidcontroller::hidcontroller( QWidget* parent,  const char* name, WFlags fl )
    : hidcontrollerbase( parent, name, fl )
{

  // signal connect
  connect(Slider, SIGNAL(valueChanged(int)), this, SLOT(sliderchange(int)));
  connect(loadButton, SIGNAL(clicked()), this, SLOT(moduleLoad()));
  connect(releaseButton, SIGNAL(clicked()), this, SLOT(moduleUnload()));
  connect(lockButton, SIGNAL(clicked()), this, SLOT(keyLock()));
  connect(unlockButton, SIGNAL(clicked()), this, SLOT(keyUnlock()));
  connect(closeButton, SIGNAL(clicked()), this, SLOT(closeClicked()));

  connect(textEdit, SIGNAL(returnPressed()), this, SLOT(textSend()));
  connect(sendButton, SIGNAL(clicked()), this, SLOT(textSend()));

  connect(clickL, SIGNAL(clicked()), this, SLOT(mouseLClicked()));
  connect(clickR, SIGNAL(clicked()), this, SLOT(mouseRClicked()));
  connect(dragL, SIGNAL(toggled(bool)), this, SLOT(mouseLDraged(bool)));
  connect(dragR, SIGNAL(toggled(bool)), this, SLOT(mouseRDraged(bool)));


  // create Menu Bar
  QPopupMenu *settings = new QPopupMenu( this );
  settings->insertItem( "Load Modules",  this, SLOT(moduleLoad()) );
  settings->insertItem( "Release Modules", this, SLOT(moduleUnload()) );
  settings->insertItem( "Key Lock", this, SLOT(keyLock()) );
  settings->insertItem( "Key Unlock", this, SLOT(keyUnlock()) );
  settings->insertSeparator();
  settings->insertItem( "Exit",  this, SLOT(closeClicked()) );

  debug = new QPopupMenu( this );
  idDebug = debug->insertItem( "debug", this, SLOT(setDebug()) );
  debug->setCheckable( TRUE );
  debug->setItemChecked(idDebug, FALSE);

  options = new QPopupMenu( this );
  idKeyMode  = options->insertItem( "Key Mode", this, SLOT(sendKeyMode()) );
  idTextMode = options->insertItem( "Send Text", this, SLOT(sendTextMode()) );
  idComMode  = options->insertItem( "Send Command", this, SLOT(sendComMode()) );
  options->setCheckable( TRUE );
  sendKeyMode();

  menuBar = new QMenuBar(this);
  menuBar->insertItem( "Settings", settings );
  menuBar->insertItem( "Options", options );
  menuBar->insertItem( "Debug", debug );
  menuBar->setFrameStyle( QFrame::Panel | QFrame::Raised );

  // init variable
  Slider->setValue(75);
  modify = 0;
  mouseButton = 0;
  closeApp = false;
  infoText->hide();

  setCaption( tr( "Hidcontroller" ) );


  /* default key data load from keydata.h */
  for(int i = 0; keyData[i][0] != 0x0000; i++){
    KeyData k;
    k.qkey = keyData[i][0];
    k.ukey = keyData[i][1];
    convKey.append(k);
  }
  
  // customized key data
  char buf[256];
  int q, u;
  fp = fopen("/home/zaurus/Settings/hidcontroller.conf", "r");
  if(fp){
    while( fgets(buf, sizeof(buf), fp) != NULL){

      if(sscanf(buf, " {%x, %x}", &q, &u) == 2){ 

	QValueList<KeyData>::Iterator it;
	bool found = false;
	for( it = convKey.begin(); it != convKey.end(); ++it ){
	  if((*it).qkey == q){
	    (*it).ukey = u;
	    found = true;
	  }
	}
	if(!found && (q != 0)){
	  KeyData k;
	  k.qkey = q;
	  k.ukey = u;
	  convKey.append(k);
	}
      }
    }
    fclose(fp);
  }

  // suspend off
  Config qpeconfig("qpe");
  if (qpeconfig.isValid()){
    qpeconfig.setGroup("Screensaver");
    if (qpeconfig.hasKey("Interval"))
      {
	intervalSuspend = qpeconfig.readNumEntry("Interval", 0);
	intervalLightOff = qpeconfig.readNumEntry("Interval_LightOff", 0);
	intervalDim = qpeconfig.readNumEntry( "Interval_Dim", 0 );
      }
  }
  toggleSuspend(false);

  // read usbd type
  QFile f("/etc/hotplug/usbd.ftype");
  QString s;
  if ( f.open(IO_ReadOnly) ) {
    QTextStream t( &f );
    s = t.readLine();
    moduleMode->setText(s);
    f.close();
  }

  // keyhelper configration load
  isKHloaded=(fopen("/home/zaurus/Settings/keyhelper_HID.xml", "r") != NULL) ? 1 : 0;
  if (isKHloaded) {
    system("qcop QPE/KeyHelper \"reload(QString)\" keyhelper_HID.xml"); //QcopEnvelope can't use. 
  }

}

/*  
 *  Destroys the object and frees any allocated resources
 */

hidcontroller::~hidcontroller()
{
   // no need to delete child widgets, Qt does it all for us

  toggleSuspend(true);

  releaseKeyboard();
  QPEApplication::ungrabKeyboard();

  if (isKHloaded)
    system("qcop QPE/KeyHelper \"reload()\""); //QcopEnvelope can't use. 

}

// only close button to quit
void hidcontroller::closeEvent(QCloseEvent *e) {
  if(closeApp)
    e->accept();
  else
    e->ignore();
}

void hidcontroller::closeClicked() {
  if(moduleMode->text() == "HID"){
    return;
  }else {
    closeApp = true;
    close();
  }
}

void hidcontroller::moduleLoad()
{
  char *command = "/home/QtPalmtop/share/hidcontroller/moduleload.sh";

  bool status = false;
  bool state = false;

  QFile f("/proc/usbd");
  QString s;
  if ( f.open(IO_ReadOnly) ) {
    QTextStream t( &f );        // use a text stream
    QString s;
    while ( !t.eof() ) {        // until end of file...
      s = t.readLine();       // line of text excluding '\n'
      if (s.find("Device status: USBD_SUSPENDED") != -1){
	status = true;
      }
      if (s.find("Device state: STATE_DEFAULT") != -1){
	state = true;
      }
    }
    f.close();
  }

  if(!status && !state){
    infoText->setText("Cable Plugged!");
    return;
  }

  if(moduleMode->text() == "net"){
    system(command);
    infoText->setText("modules loaded");
    moduleMode->setText("HID");
  }else { 
    infoText->setText("Can't load");
  }
}

void hidcontroller::moduleUnload()
{
  char *command = "/home/QtPalmtop/share/hidcontroller/moduleunload.sh";

  bool status = false;
  bool state = false;

  QFile f("/proc/usbd");
  QString s;
  if ( f.open(IO_ReadOnly) ) {
    QTextStream t( &f );        // use a text stream
    QString s;
    while ( !t.eof() ) {        // until end of file...
      s = t.readLine();       // line of text excluding '\n'
      if (s.find("Device status: USBD_SUSPENDED") != -1){
	status = true;
      }
      if (s.find("Device state: STATE_DEFAULT") != -1){
	state = true;
      }
    }
    f.close();
  }

  if(!status && !state){
    infoText->setText("Cable Plugged!");
    return;
  }

  if(moduleMode->text() == "HID"){
    system(command);
    infoText->setText("modules unloaded");
    moduleMode->setText("net");
  }else {
    infoText->setText("Can't Release");
  }
}

void hidcontroller::keyUnlock()
{
  keyMode->setText("Unlock");
  textEdit->setFocusPolicy( StrongFocus );
  sendButton->setFocusPolicy( StrongFocus );
  releaseKeyboard();
  QPEApplication::ungrabKeyboard();
}

void hidcontroller::keyLock()
{
  QPEApplication::grabKeyboard();
  textEdit->setFocusPolicy( NoFocus );
  sendButton->setFocusPolicy( NoFocus );
  grabKeyboard();
  keyMode->setText("Lock");
}

void hidcontroller::sliderchange( int val )
{
  QString str;
  str.sprintf("%2.2f",(100 - val)/25.0);
  accel->setText(str);
}

void hidcontroller::toggleSuspend(bool enable)
{
    QCopEnvelope e("QPE/System", "setScreenSaverIntervals(int,int,int)" );
    if (enable)
    {
      e << intervalDim << intervalLightOff << intervalSuspend;
    } 
    else 
      {
	e << 0 << 0 << 0;
      }
}

void hidcontroller::setDebug()
{
  debug->setItemChecked(idDebug, !(debug->isItemChecked(idDebug)) );
  if(debug->isItemChecked(idDebug))
    infoText->show();
  else
    infoText->hide();
}

/*********************************************************************
	mouse control
**********************************************************************/

void hidcontroller::mouseMoveEvent( QMouseEvent *e)
{
  QString str;
  int dx,dy;
  int mag;

  if(moduleMode->text() != "HID")
    return;

  if(mouseNotValid)
    return;

  dx = e->globalX() - x;
  dy = e->globalY() - y;

  mag = 100-Slider->value();

  str.sprintf("d %02x %02x, %02x: %d %d slider %d", mouseButton,
	      (char)(dx * mag / 25), (char)(dy * mag / 25),dx*mag,dx*mag/25, mag);


  x = e->globalX();
  y = e->globalY();

    infoText->setText(str);

  fp = fopen("/dev/hidout11", "w");
  if(fp){
    fprintf(fp, "%03x %03x %03x\n", mouseButton, 
	    (char)(dx * mag / 25), (char)(dy * mag / 25));
    
    fclose(fp);
  }
}

void hidcontroller::mousePressEvent( QMouseEvent *e)
{
  x = e->globalX();
  y = e->globalY();

  if(x < 100 || x > 550 || y < 20 || y > 250){
    mouseNotValid = true;
  }else{
    mouseNotValid = false;
  }

}

void hidcontroller::mouseLClicked()
{

  mouseButton |= 0x01;

  fp = fopen("/dev/hidout11", "w");
  if(fp){
    fprintf(fp, "%03x 00 00\n", mouseButton);
      fclose(fp);
  }

  mouseButton &= ~0x01;

  fp = fopen("/dev/hidout11", "w");
  if(fp){
    fprintf(fp, "%03x 00 00\n", mouseButton);
      fclose(fp);
  }

  return;
}

void hidcontroller::mouseRClicked()
{

  mouseButton |= 0x02;

  fp = fopen("/dev/hidout11", "w");
  if(fp){
    fprintf(fp, "%03x 00 00\n", mouseButton);
      fclose(fp);
  }

  mouseButton &= ~0x02;

  fp = fopen("/dev/hidout11", "w");
  if(fp){
    fprintf(fp, "%03x 00 00\n", mouseButton);
      fclose(fp);
  }

  return;
}

void hidcontroller::mouseLDraged(bool press)
{

    if(press){
      mouseButton |= 0x01;
    }else {
      mouseButton &= ~0x01;
    }
    fp = fopen("/dev/hidout11", "w");
    if(fp){
      fprintf(fp, "%03x 00 00\n", mouseButton);
      fclose(fp);
    }
    return;

}

void hidcontroller::mouseRDraged(bool press)
{

    if(press){
      mouseButton |= 0x02;
    }else {
      mouseButton &= ~0x02;
    }
    fp = fopen("/dev/hidout11", "w");
    if(fp){
      fprintf(fp, "%03x 00 00\n", mouseButton);
      fclose(fp);
    }
    return;

}

/*********************************************************************
	key control
**********************************************************************/

void hidcontroller::QueueKey( QKeyEvent *e, int pressed )
{
  QString str;
  int udata;

  int scancode = e->key();
  unsigned int data[6];
  unsigned char modifyonce;

  char led[10];

  if(moduleMode->text() != "HID")
    return;

  modifyonce = 0;

  QValueList<KeyData>::Iterator it;
  udata = 0;
  for( it = convKey.begin(); it != convKey.end(); ++it ){
    if((*it).qkey == scancode){
      udata = (*it).ukey;
    }
  }

  if( (udata >> 12) & 0x0f ) {					// mouse button
    if(pressed){
      mouseButton |= ((udata >>12) &0x0f);
    }else {
      mouseButton &= ~((udata >>12) &0x0f);
    }
    fp = fopen("/dev/hidout11", "w");
    if(fp){
      fprintf(fp, "%03x 00 00\n", mouseButton);
      fclose(fp);
    }
    return;

  }else if( ((udata >> 8) & 0x0f) && ((udata & 0xff) == 0x00)) { // modify key
    if(pressed){
      modify |= ((udata >> 8) & 0x0f);
    }else {
      modify &= ~((udata >> 8) & 0x0f);
    }


  }else {							// normal key
    if(pressed){
      sendData.append(udata & 0xff);
      modifyonce = ((udata >> 8) & 0x0f);
    }else {
      sendData.remove(udata & 0xff);
      modifyonce = 0x0;
    }
  }
  
  QValueList<int>::Iterator ite;
  int i=0;
  for(i = 0; i < 6; i++){
    data[i] = 0;
  }
  i = 0;
  for(ite = sendData.begin(); ite != sendData.end() && i < 6; ++ite ){
    data[i++] = (*ite);
  }
  
  fp = fopen("/dev/hidout2", "w");
  if(fp){
    fprintf(fp, "%03x 00 %03x %03x %03x %03x %03x %03x\n", (modify | modifyonce), data[0], data[1], data[2], data[3], data[4], data[5]);
    fclose(fp);
  }

  fp = fopen("/dev/hidout0", "r");
  if(fp){
    fgets(led, 5, fp);
    fclose(fp);
  }
  
  if(led[1] & 0x01){
    textNumLock->setText("NUM");
  }else {
    textNumLock->setText("");
  }  

  if(led[1] & 0x02){
    textCapsLock->setText("CAPS");
  }else {
    textCapsLock->setText("");
  }  

  if(led[1] & 0x04){
    textScrlLock->setText("SCRL");
  }else {
    textScrlLock->setText("");
  }  

  str.sprintf("key %04x Press %x DATA:%02x 00 %02x %02x %02x %02x %02x %02x : LED %2x"
	      ,scancode, pressed, (modify | modifyonce), 
	      data[0], data[1], data[2], data[3], data[4], data[5], led[1] & 0x0f);
  infoText->setText(str);
  
}

static unsigned int char_tbl[] =
  {
    0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x002a,0x002b,0x0028,0x0000,0x0000,0x0000,0x0000,0x0000,
    0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0029,0x0000,0x0000,0x0000,0x0000,
    0x002c,0x021e,0x0234,0x0220,0x0221,0x0222,0x0224,0x0034,0x0226,0x0227,0x0225,0x022e,0x0036,0x002d,0x0037,0x0038,
    0x0027,0x001e,0x001f,0x0020,0x0021,0x0022,0x0023,0x0024,0x0025,0x0026,0x0233,0x0033,0x0236,0x002e,0x0237,0x0238,
    0x021f,0x0204,0x0205,0x0206,0x0207,0x0208,0x0209,0x020a,0x020b,0x020c,0x020d,0x020e,0x020f,0x0210,0x0211,0x0212,
    0x0213,0x0214,0x0215,0x0216,0x0217,0x0218,0x0219,0x021a,0x021b,0x021c,0x021d,0x002f,0x0031,0x0030,0x0223,0x022d,
    0x0034,0x0004,0x0005,0x0006,0x0007,0x0008,0x0009,0x000a,0x000b,0x000c,0x000d,0x000e,0x000f,0x0010,0x0011,0x0012,
    0x0013,0x0014,0x0015,0x0016,0x0017,0x0018,0x0019,0x001a,0x001b,0x001c,0x001d,0x022f,0x0231,0x0230,0x0234,0x004c
  };

typedef struct alt_char_table
{
  char *text;
  unsigned int code;
} alt_char_table;

static alt_char_table alt_char_tbl[] =
  {
    {"[ctrl]",	0x0100},
    {"[alt]",	0x0400},
    {"[spc]",	0x002c},
    {"[caps]",	0x0039},
    {"[tab]",	0x002b},
    {"[esc]",	0x0029},
    {"[ret]",	0x0028},
    {"[shift]", 0x0200},
    {"[bs]",	0x002a},
    {"[ins]",	0x0049},
    {"[del]",	0x004c},
    {"[home]",	0x004a},
    {"[end]",	0x004d},
    {"[pgup]",	0x004b},
    {"[pgdn]",	0x004e},
    {"[up]",	0x0052},
    {"[down]",	0x0051},
    {"[left]",	0x0050},
    {"[right]", 0x004f},
    {"[numlk]", 0x0053},
    {"[prtsc]", 0x0046},
    {"[scrlk]", 0x0047},
    {"[pause]", 0x0048},
    {"[kp/]",	0x0054},
    {"[kp*]",	0x0055},
    {"[kp+]",	0x0057},
    {"[kp-]",	0x0056},
    {"[kpret]", 0x0058},
    {"[kp.]",	0x0063},
    {"[kp0]",	0x0062},
    {"[kp1]",	0x0059},
    {"[kp2]",	0x005a},
    {"[kp3]",	0x005b},
    {"[kp4]",	0x005c},
    {"[kp5]",	0x005d},
    {"[kp6]",	0x005e},
    {"[kp7]",	0x005f},
    {"[kp8]",	0x0060},
    {"[kp9]",	0x0061},
    {"[f1]",	0x003a},
    {"[f2]",	0x003b},
    {"[f3]",	0x003c},
    {"[f4]",	0x003d},
    {"[f5]",	0x003e},
    {"[f6]",	0x003f},
    {"[f7]",	0x0040},
    {"[f8]",	0x0041},
    {"[f9]",	0x0042},
    {"[f10]",	0x0043},
    {"[f11]",	0x0044},
    {"[f12]",	0x0045},

  };

void hidcontroller::textSend()
{
  QString text;
  QStringList comList;
  unsigned int data[6];

  if(moduleMode->text() != "HID")
    return;
  
  text = textEdit->text();

  if(text.isEmpty())
    return;

  if( options->isItemChecked(idTextMode) ){

    fp = fopen("/dev/hidout1", "w");
    if(fp){
      fprintf(fp, "%s", text.latin1());
    fclose(fp);

    }
  }else if( options->isItemChecked(idComMode) ){

    comList = QStringList::split(" ", text);

    for ( QStringList::Iterator it = comList.begin(); it != comList.end(); ++it ) {
      uint pos = 0;
      unsigned char modifyonce = 0;

      const char *ch = (*it).latin1();
      for(uint i = 0; i < 6; i++){
	if(pos < strlen(ch)){
	  bool found = false;
	  for(uint j = 0; j < sizeof(alt_char_tbl)/sizeof(alt_char_table); j ++){
	    if(pos + strlen(alt_char_tbl[j].text) > strlen(ch)){
	      continue;
	    }
	    if(strncmp((ch+pos), alt_char_tbl[j].text, strlen(alt_char_tbl[j].text)) == 0){
	      data[i]     = (alt_char_tbl[j].code) & 0xff;
	      modifyonce |= (alt_char_tbl[j].code >> 8) & 0xff;
	      pos += strlen(alt_char_tbl[j].text);
	      found = true;
	      break;
	    }
	  }

	  if(!found){
	    data[i]     = (char_tbl[(*(ch + pos)) & 0x7f]) & 0xff;
	    modifyonce |= (char_tbl[(*(ch + pos)) & 0x7f] >> 8) & 0xff;
	    if( ((*(ch + pos)) == '[') && ((*(ch + pos + 1)) == '[') ) {
	      pos ++;
	    }
	    pos++;
	  }
	}else{
	  data[i] = 0;
	}
      }

      fp = fopen("/dev/hidout2", "w");
      if(fp){
	fprintf(fp, "%03x 00 %03x %03x %03x %03x %03x %03x\n",
		modifyonce, data[0], data[1], data[2], data[3], data[4], data[5]);
	fclose(fp);
      }

    }

    fp = fopen("/dev/hidout2", "w");
    if(fp){
      fprintf(fp, "00 00 00 00 00 00 00 00\n");
      fclose(fp);
    }
  }  
}

void hidcontroller::sendKeyMode()
{
  sendButton->hide();
  textEdit->hide();
  options->setItemChecked(idKeyMode, TRUE);
  options->setItemChecked(idTextMode, FALSE);
  options->setItemChecked(idComMode, FALSE);
  keyLock();
}

void hidcontroller::sendTextMode()
{
  sendButton->show();
  textEdit->show();
  sendButton->setText("SendText");
  options->setItemChecked(idKeyMode, FALSE);
  options->setItemChecked(idTextMode, TRUE);
  options->setItemChecked(idComMode, FALSE);
  keyUnlock();		
}

void hidcontroller::sendComMode()
{
  sendButton->show();
  textEdit->show();
  sendButton->setText("SendCom");
  options->setItemChecked(idKeyMode, FALSE);
  options->setItemChecked(idTextMode, FALSE);
  options->setItemChecked(idComMode, TRUE);
  keyUnlock();
}
